#!/usr/bin/perl -w
#
# higherLabels.pl v1.1
# Last modified February 7, 2005 12:39
# (c) Olaf R.P. Bininda-Emonds
#
# Input:
#   1) A nexus-formatted tree file
#	2) A tab-delimited file containing species names (in first column) and the higher-level taxa they belong to
#	   (remaining columns, in order of increasing inclusiveness)
#
# Output:
#	1) treefile_labeled.tre: A nexus-formatted treefile with higher-level taxon names assigned to the most-inclusive node(s)
#	   possible as a node label. Non-monophyletic higher-level taxa are accounted for and are labeled sequentially together
#	   with an asterisk.
#	2) treefile_labeled.txt: A text file summarizing the labels given to the nodes in (1)
#	2) unmatchedFamilies.txt: A text file listing any higher-level taxa
#
# Usage: higherLabels.pl -c<filename> -t<filename> [-v] [-h]
#   options: -c<filename> = tab-delimited file containing species names (in first column) and the higher-level
#							taxa they belong to (remaining columns, in order of increasing inclusiveness)
#            -t<filename> = treefile containing trees in NEXUS format
#			 -v = verbose mode (not recommended!)
#            -h = print this message and quit

use strict;

# Set defaults and clear global variables
	# Taxonomy variables
		my $taxFile;
		my (@taxonomy, %taxonTaxonomy);
		
	# Tree variables
		my ($treeRead, $treeWrite, $textWrite, $dataSource);
		my %taxonTree;
#			$treeRead = "mammalST.tre";
		my @allTrees;

	# Global variables
		my $verbose = 0;
		my $debug = 0;

for (my $i = 0; $i <= $#ARGV; $i++)
	{
	if ($ARGV[$i] =~ /-c([\w+\.?\w+?]*)/)
		{
		$taxFile = $1;
		}
	elsif ($ARGV[$i] =~ /-t([\w+\.?\w+?]*)/)
		{
		$treeRead = $1;
		($dataSource) = split(/\./,$treeRead,0);
			$treeWrite = $dataSource."_labeled.tre";
			$textWrite = $dataSource."_labeled.txt";
		}
	elsif ($ARGV[$i] =~ /-v/)
		{
		$verbose = 1;
		}
	elsif ($ARGV[$i] =~ /-x/)
		{
		$debug = 1;
		}
	elsif ($ARGV[$i] eq "-h")
		{
		print "Usage: higherLabels.pl -c<filename> -t<filename> [-v] [-h]\n";
		print "Options: -c<filename> = tab-delimited file containing species names (in first column) and the higher-level\n";
		print "                        taxa they belong to (remaining columns, in order of increasing inclusiveness)\n";
		print "         -t<filename> = treefile containing trees in NEXUS format\n";
		print "         -v = verbose mode (not recommended!)\n";
		print "         -h = print this message and quit\n";
		exit(0);
		}
	else
		{
		print "Don't understand argument: $ARGV[$i]\n";
		print "Usage: higherLabels.pl -c<filename> -t<filename> [-v] [-h]\n";
		exit(1); 
		}
	}

die "ERROR: Must specify name of treefile and/or file listing higher-level clades for species.\n" if (not defined $treeRead or not defined $taxFile);

# Read in file assigning species names to families and store
	print "Reading in user-input taxonomy ...\n";
	my $taxZero = time;
	setLineBreak("$taxFile");
	open (TAX, "<$taxFile") or die "Cannot open file assigning species to families, $taxFile.\n";
		while (<TAX>)
			{
			next unless ($_);
			chomp;
			push @taxonomy, $_;
			
			my (@taxLine) = split(/\t/, $_);	# Store species that are present in taxonomy
				$taxonTaxonomy{shift(@taxLine)} = 1;
			}
	close TAX;
	die "ERROR: No data read in from user-input taxonomy\n" unless (@taxonomy);
	printf "\tTime taken to read in taxonomy of %s species: %s seconds\n", scalar(@taxonomy), time - $taxZero;
	
# Open output files and print header for treefile
	open (TREE, ">$treeWrite") or die "Cannot open output file $treeWrite to write tree description(s).\n";
		print TREE "#NEXUS\n\nbegin trees;\n";
	open (TEXT, ">$textWrite") or die "Cannot open output file $textWrite to write family names.\n";
	open (NAME, ">unmatchedFamilies.txt") or die "Cannot open output file unmatchedFamilies.txt to write unmatched higher level names.\n";

# Read in supertree and store information
	print "\nReading in tree description(s) ...\n";
	my @treeList = treeRead($treeRead);
		my $treeCount = 0;
	
	foreach my $superTree (@treeList)
		{
		$superTree =~ s/:\d+\.?\d*e?\-?\d*//g;	# Remove any branch length information
		
		# Basic check of tree structure
			my $numSTclade = ($superTree =~ tr/(//);
				if ($numSTclade < 2)
					{
					print "\tERROR: Tree does not have enough informative nodes ($numSTclade clades); skipping\n";
					}
		
		my $treeZero = time;
		$treeCount++;
		print "\tProcessing tree $treeCount ...\n";
		print "\t\tTree description: $superTree\n" if ($verbose);
		
		# Mark taxa present on tree
			my $taxTree = $superTree;
				$taxTree =~ s/\(//g;
				$taxTree =~ s/\)//g;
				
			my @treeTaxonList = split(',', $taxTree);
				$taxonTree{$_} = 1 foreach (@treeTaxonList);
			
		# Process user-input taxonomy, matching taxon set on tree
			print "\t\tAdjusting user-input taxonomy to match taxon set...\n";
			my $taxZero = time;
			my (@familyNames, %familyPresent, %familyDepth, %familyMembership, %familySize, %familySizePresent, %familyID, %familyMatch, %currentClade, %familyCount);
			foreach my $line (@taxonomy)
				{
				my (@taxLine) = split(/\t/, $line);
				my $species = shift(@taxLine);
				next unless (defined $taxonTree{$species});	# Only process species present on tree

				for (my $level = 0; $level < scalar(@taxLine); $level++)	# Process higher-level names and assign to species
					{
					my $family = $taxLine[$level];
					next if ($family eq "");
					unless ($familyPresent{$family})
						{
						push @familyNames, $family;
						$familyPresent{$family} = 1;
						$familyDepth{$family} = $level;
						}

					push @ { $familyMembership{$family} }, $species;
					}

			@familyNames = sort { $familyDepth{$b} <=> $familyDepth{$a} } keys %familyDepth;
			
			# Process each family
				foreach my $family (@familyNames)
					{
					$familySize{$family} = scalar (@ { $familyMembership{$family} });
					$familySizePresent{$familySize{$family}} = 1;
					@ { $familyMembership{$family} } = sort(@ { $familyMembership{$family} });
					print "\t\t\t$family\t$familySize{$family} species\n" if ($verbose);
					}
				}
			printf "\t\t\tTime taken: %s seconds\n", time - $taxZero;

		# Parse supertree to get clade memberships and sizes
			print "\t\tParsing supertree description ...\n";
			my $parseZero = time;
			my (%ParentClade, %daughterClade, %CladeDescrip, %cladeMembership, %CladeSupport, %STcladeSize, %open, %start, %stop, %present);
			my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
			my $OldCurrent;
			while ($position < length($superTree))
				{
				my $offset = 1;
				if (substr($superTree, $position, 1) eq "(")		#New clade; go one level deeper
					{
					$CladeID++;
					$ParentClade{$CladeID} = $CurrentClade;
					$CurrentClade = $CladeID;
					$open{$CurrentClade} = 1;
					$start{$CladeID} = $position;
					}
				elsif (substr($superTree, $position, 1) eq ")")	#Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					push @{$daughterClade{$CurrentClade} }, $OldCurrent;
					$stop{$OldCurrent} = $position;
					$CladeDescrip{$OldCurrent} = substr($superTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
					$CladeDescrip{$OldCurrent} =~ s/:\d+\.?\d*//g;	# Remove branch length information
					}
				elsif (substr($superTree, $position, 1) eq ",")
					{
					# Do nothing
					}
				else	# Determine membership of clade
					{
					until (substr($superTree, $position+$offset, 1) eq "(" or substr($superTree, $position+$offset, 1) eq ")" or substr($superTree, $position+$offset, 1) eq "," or substr($superTree, $position+$offset, 1) eq ":")
						{
						$offset++;
						}
					my $taxon = substr($superTree, $position, $offset);
					print "\t\t\tWARNING: $taxon does not appear in user-input taxonomy\n" if (not defined $taxonTaxonomy{$taxon});
					for (my $clade = 1; $clade <= $CladeID; $clade++)	# Assign all open clades to taxon and increment size of open clades
						{
						if ($open{$clade})
							{
							push @{ $cladeMembership{$clade} }, $taxon;
							$STcladeSize{$clade}++;
							}
						}
					$position += $offset-1;
					}
				$position++;
				}
			printf "\t\t\tTime taken: %s seconds\n", time - $parseZero;
		
		# Get family names for supertree clades
			print "\t\tGetting higher level names for supertree clades ...\n";
			my $labelZero = time;
			$STcladeSize{"0"} = 0;
			print NAME "Tree $treeCount\n";
			foreach my $family (@familyNames)
				{
				if ($familySize{$family} == 1)	# Discard monotypic families; no matching node possible
					{
					print NAME "\t$family ($familySize{$family})\n";
					next;
					}
				print "\tLevel $familyDepth{$family}: $family\n" if ($debug);
				for (my $clade = $numSTclade; $clade > 0; $clade--)	# Process clades in reverse order to go from less to more inclusive
					{
					next if ($STcladeSize{$clade} > $familySize{$family});
#					next if ($familyID{$clade});
					# Compute union of membership arrays for family and clade
						my (@union, @intersection, @difference);
						my %count;
						foreach my $element (@{ $familyMembership{$family}}, @{ $cladeMembership{$clade}})
							{
							$count{$element}++;
							}
						foreach my $element (keys %count)
							{
							push @union, $element;
							push @{ $count{$element} > 1 ? \@intersection : \@difference }, $element;
							}
					if (scalar(@union) <= $familySize{$family})	# Accounts for missing taxa
						{
						$familyMatch{$family} = 1;
						$familyID{$clade} .= $family."_";	# unless (defined $familyID{$ParentClade{$clade}} and $familyID{$ParentClade{$clade}} eq $family);
						foreach my $daughter (@{$daughterClade{$clade} })	# Correction to apply name to most inclusive clade only
							{
							if (defined $familyID{$daughter} and $familyID{$daughter} =~ m/$family\_/)
								{
								$familyID{$daughter} =~ s/$family\_//;
								}
							}
						}
					}
				unless ($familyMatch{$family})
					{
					print NAME "\t$family ($familySize{$family})\n";
					}
				}
			printf "\t\t\tTime taken: %s seconds\n", time - $labelZero;
		
		# Print labelled tree to treefile
			$position = $CurrentClade = $CladeID = 0;
			undef %ParentClade;
			undef %open;
			my $newST;
			undef $OldCurrent;
			while ($position < length($superTree))
				{
				my $offset = 1;
				if (substr($superTree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$CladeID++;
					$ParentClade{$CladeID} = $CurrentClade;
					$CurrentClade = $CladeID;
					$newST .= "(";
					}
				elsif (substr($superTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					$newST .= ")";
					if ($OldCurrent == 0)
						{
						$newST .= "root";
						}
					else
						{
						if ($familyID{$OldCurrent})
							{
#							print "\n$OldCurrent\n\t$familyID{$OldCurrent}\t" if ($debug);
							$familyID{$OldCurrent} =~ s/\_$//;
#							print "\t$familyID{$OldCurrent}\n" if ($debug);
							my $label;
							my @familyList = split(/_/, $familyID{$OldCurrent});
							foreach my $realName (@familyList)
								{
#$label .= "_$realName";
#next;
#								print "\t\t$realName\t" if ($debug);
								$familyCount{$realName}++;
								my $name = $realName;
								if ($familyCount{$realName} > 1)
									{
									$name .= $familyCount{$realName};
									$familyID{$OldCurrent} =~ s/$realName/$name/;
									}
#								print "$name\n" if ($debug);
								$label .= "_$name";
								$label .= "*" if ($STcladeSize{$OldCurrent} != $familySize{$realName})
								}
							$label =~ s/^_//;
#							print "\t$label\n" if ($debug);
							$newST .= "'$label'";
							}
						print "\t$OldCurrent: $familyID{$OldCurrent}\n" if (defined $familyID{$OldCurrent} and $debug);
						}
					}
				else
					{
					$newST .= substr($superTree, $position, 1);
					}
				$position++;
				}
			print TREE "\tTREE $dataSource = $newST;\n";

		# Print list of all nodes with family names if applicable
			print TEXT "Nodes for tree $treeCount:\n";
			for (my $clade = 1; $clade <= $CladeID; $clade++)
				{
				print TEXT "\tNode " . $clade . ":";
#				print "\n$clade" if ($debug);
				if (defined $familyID{$clade})
					{
#					print "\t$familyID{$clade}" if ($debug);
					foreach my $name (split(/_/, $familyID{$clade}))
						{
						print TEXT "\t$name";
						(my $realName = $name) =~ s/\d+$//;
#						print "\t$realName" if ($debug);
						print TEXT "*" if ($STcladeSize{$clade} != $familySize{$realName});
						}
					}
				print TEXT "\n";
				}
		printf "\t\tTotal time taken for tree $treeCount: %s seconds\n", time - $treeZero;
		}

# Print footer to treefile and close; also close node file
	print TREE "end;\n";
	close TREE;
	close TEXT;
	close NAME;

exit(0);

# Subroutines used in the script

sub treeRead
	{
	my $treefile = shift;
	my $block = 0;
	my @AllTrees;
	setLineBreak($treefile);
	open (fTREES, "<$treefile") or die "Cannot open treefile $treefile\n";
		my ($treeDescrip, $treeFlag, $treeCount);
		my (@taxonList, %taxonPresent);
		my $treeBlock = 0;
		my (%TTable, $translation, $translateFlag, $maxTaxonID, $transText);
		
		while (<fTREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			if ($_ =~ m/begin trees;/i)		# Clear variables for each tree block
				{
				undef %TTable;
				$translation = $translateFlag = $treeFlag = $maxTaxonID = $treeCount = 0;
				$treeBlock++;
				}
			if ($_ =~ m/translate/i)	#Read translation table of taxon names (if any)
				{
				$translateFlag = $translation = 1;
				$transText = "";
				print "Reading translation table ...\n" if ($verbose);
				}
			if ($translateFlag)
				{
				unless ($_ =~ m/translate/i)
					{
					$transText .= $_;
					if ($_ =~ m/;/)
						{
						$translateFlag = 0;
						$transText =~ s/;$//;	#Strip trailing semi-colon
						my @transEntries = split(/,/, $transText);
						foreach my $entry (@transEntries)
							{
							$entry =~ s/^\s+//;		#Strip (any) leading white space
							my ($taxonID, $taxonName) = split(/\s/,$entry);
							$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
							$taxonName =~ s/\s+//g;
							$TTable{$taxonID} = $taxonName;
							}
						}
					}
				}
			if ($_ =~ /tree/i and $_ =~ /=/)
				{
				unless ($_ =~ /^>/)		#Ignore (any) commented lines
					{
					$treeDescrip = "";
					unless ($block == 0 and $treeBlock > 1)
						{
						$treeFlag = 1;
						}
					}
				}
			if ($treeFlag)
				{
				$treeDescrip .= $_;
				if ($_ =~ m/;/)
					{
					$treeCount++;
					print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
					print "\tTranslating taxon names ...\n" if ($verbose and $translation);
					$treeFlag = 0;
					my $position = 0;
					until (substr($treeDescrip, $position, 1) eq "(")		#Avoid anything before tree description proper
						{
						$position++;
						}
					my $tree = substr($treeDescrip, $position);
						$tree =~ s/;\s*$//;
						$tree =~ s/\s+//g;
					my ($name) = (split /=/, $treeDescrip, 0);		#Get name of tree, if any
						$name =~ s/\s*\w*tree\s+//i;
						$name =~ s/\s*\*\s*//i;
						$name =~ s/\s+$//g;
#						$treeNames{$treeCount} = $name;
					my $openCount = ($tree =~ tr/(//);
					my $closedCount = ($tree =~ tr/)//);
					if ($openCount == $closedCount)		#Tree is balanced; process
						{
						$position = length($tree) - 1;
						while ($position >= 0)
							{
							my $IDlength = 1;
							if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
								{
								$position--;
								}
							else	# Read in taxonID
								{
								until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
									{
									$position--;
									$IDlength++;
									}
								unless (substr($tree, $position, 1) eq ":")		# Avoid reading branch lengths as taxon names
									{
									my $readName = substr($tree, $position+1, $IDlength-1);
									if ($translation)	# Translate taxon names
										{
										if ($TTable{$readName})
											{
											substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		#Translate name
											unless ($taxonPresent{$TTable{$readName}})
												{
												push @taxonList, $TTable{$readName};
												$taxonPresent{$TTable{$readName}} = 1;
												}
											}
										else
											{
											print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
											substr($tree, $position+1, $IDlength-1) = "xxx";		#Translate name
											unless ($taxonPresent{"xxx"})
												{
												push @taxonList, "xxx";
												$taxonPresent{"xxx"} = 1;
												}
											}
										}
									else	#Enter name as read from tree, unless already present
										{
										unless ($taxonPresent{$readName})
											{
											push @taxonList, $readName;
											$taxonPresent{$readName} = 1;
											}
										}
									}
								}
							}
						$tree =~s /\)\w+/\)/g;	# Remove any named nodes
						push @AllTrees, $tree;
					print "\tTree description:\n\t\t$tree\n" if ($verbose);
						}
					else	#Tree is unbalanced; reject
						{
						print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
						}
					}
				}
			}
	close fTREES;
	die "\nERROR: No trees read in from file $treefile\n" if (not @AllTrees);
	
	return @AllTrees;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "DOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "Mac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "Unix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
#	v1.1 February 7, 2005
#		- removed restriction that user-input taxonomy now assumed to be correct;
# 		  must no longer match a reference taxonomy
#		- user-input taxonomy now pruned to match taxon set of tree (* taxa are
#		  therefore non-monophyletic only, not possible incomplete as well)
#		- added ability to process mutliple trees
#
#	v1.0 July 27, 2004
#		- initial release
